
# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# Many thanks to Blender Artists, Mano Wii and Elegant Artz.

bl_info = {
  'name':'NP Float Box 007',
  'author':'Okavango and the ba community',
  'version':(0,0,7),
  'blender':(2,75,0),
  'location':'View3D',
  'description':'Create a box with three clicks',
  'category':'3D View'}

import bpy
import mathutils, bmesh, math
from math import sin, cos, tan, atan, degrees, radians, asin, acos
import bgl
import blf
from mathutils import Vector, Matrix
from bpy_extras import view3d_utils

# Defining the main class - the macro:
class NPFloatBox007(bpy.types.Macro):
    bl_idname = 'np.float_box_007'
    bl_label = 'NP Float Box 007'
    bl_options = {'REGISTER','UNDO'}
    
# Defining the storage class that will serve as a varable-bank for exchange among the classes. 
#Later, this bank will recieve more variables with their values for safe keeping, as the program goes on:
class Storage:   
    z = 1.0
    use_snap = None    
    snap_element = None
    snap_target = None

# Defining the first graphical set that will be shown during first phase and updated along:    
def draw_callback1_px(self, context):    
    print('trail'+ str(len(self.trail)))
        
    # Calculating the 3d points where cursor ray hits the ground:
    region = context.region
    rv3d = context.region_data
    co2d = self.co2d   
    view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, co2d)
    ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, co2d)
    ray_target = ray_origin + view_vector * 1000
    

    center = Vector((0,0,0))
    normal = Vector((0.0, 0.0, 1.0))
    co3dv = mathutils.geometry.intersect_line_plane(ray_origin, ray_target, center, normal, False)
    co3d = co3dv.to_tuple(4)    
    print('co3d' ,co3d)
    print('raycast', context.scene.ray_cast(ray_origin, ray_target))
    
    # Storring the 3d points into lists for drawing the base: 
    if self.mode == 1:
        self.trail.append((co3d))
        print('1')
    if self.mode == 2:
        self.points.append((co3d))
        self.up[0] = co3d
        self.down[0] = co3d
        print('2')
        
    # Correcting some persistent mistake that appears, for which i don't know the reason: 
    if len(self.points) > 1:
        a = [(0,0,0)]
        a[0] = self.points[0]
        self.points = a
        print('oops!')
                   
    #These prints are my control in the process of writing:
    print('points',self.points)           
    print('up',self.up)
    print('down',self.down)

    # Adjusting the lists for the base while the mouse is still moving:
    j = len(self.trail)    
    if len(self.points) == 1:        
        xupydown = self.points[0]
        xup = xupydown[0]
        ydown = xupydown[1]         
        i = 1
        for x, y, z in self.trail:
            if i == j:
                #print(self.trail[j-1])
                self.up[1] = (xup, y, 0)
                self.down[1] = (x, ydown, 0)               
                self.up[2] = self.trail[j-1]
                self.down[2] = self.trail[j-1]
            i = i + 1
    
    #print(len(self.up))
    #print(len(self.down))
    
    # Before they get 2d flattened, recording the lists in 3d for the latter mesh generation:
    self.base[0] = self.up[0]
    self.base[1] = self.up[1]
    self.base[2] = self.up[2]
    self.base[3] = self.down[1]
    #print('self.base', self.base)
        
    # Drawing the base in 2d on the screen, from 3d points:
    bgl.glEnable(bgl.GL_BLEND)
    # First color is for lighter themes, second for default and darker themes:       
    #bgl.glColor4f(1.0, 1.0, 1.0, 1.0)
    bgl.glColor4f(1.0, 1.0, 1.0, 0.6)
    bgl.glLineWidth(1)    
    up2dlist = []
    down2dlist = []
    bgl.glBegin(bgl.GL_LINE_STRIP) 
    for co in self.up:
        print('co', co)
        up2d = view3d_utils.location_3d_to_region_2d(region, rv3d, co)
        bgl.glVertex2f(*up2d)
        up2dlist.append(up2d.to_tuple(2))
    #print(up2dlist)      
    for co in self.down:
        down2d = view3d_utils.location_3d_to_region_2d(region, rv3d, co)        
        bgl.glVertex2f(*down2d)
        down2dlist.append(down2d.to_tuple(2))
    #print(down2dlist)
    bgl.glEnd()
    
    # Filling the base with a transparent color:
    # First color is for lighter themes, second for default and darker themes:      
    #bgl.glColor4f(1.0, 1.0, 1.0, 0.3)
    bgl.glColor4f(1.0, 1.0, 1.0, 0.2)        
    bgl.glBegin(bgl.GL_TRIANGLE_FAN)
    for co in self.up:
        up2d = view3d_utils.location_3d_to_region_2d(region, rv3d, co)
        bgl.glVertex2f(*up2d)
    bgl.glEnd()    
    bgl.glBegin(bgl.GL_TRIANGLE_FAN)
    for co in self.down:
        down2d = view3d_utils.location_3d_to_region_2d(region, rv3d, co)
        bgl.glVertex2f(*down2d)
    bgl.glEnd()
    
    # Calculating the dimensions on the sides of the base:
    uploc = []
    uploc.append((up2dlist[1][0]+up2dlist[2][0])/2)
    uploc.append((up2dlist[1][1]+up2dlist[2][1])/2)    
    downloc = []
    downloc.append((down2dlist[1][0]+down2dlist[2][0])/2)
    downloc.append((down2dlist[1][1]+down2dlist[2][1])/2)    
    lenxyz = (mathutils.Vector(self.up[2]) - mathutils.Vector(self.up[0]))
    lenx = str(abs(round(lenxyz[0],2)))
    leny = str(abs(round(lenxyz[1],2))) 
    
    # Drawing the small icon near the cursor and the dimensions:
    # First color is for lighter themes, second for default and darker themes:
    #bgl.glColor4f(1.0, 1.0, 1.0, 1.0)
    bgl.glColor4f(1.0, 1.0, 1.0, 0.8)        
    if self.up[0] == (0,0,0) and self.up[1] == (0,0,0) and self.up[2] == (0,0,0):
        cur = [[18, 30], [15, 32], [15, 36], [18, 34], [18, 30], [21, 32], 
        [21, 36], [18, 34], [15, 36], [18, 38], [21, 36]]    
        scale = 2.0
        for co in cur:    
            co[0] = round((co[0] * scale),0) -13 + co2d[0]
            co[1] = round((co[1] * scale),0) -65 + co2d[1]        
        bgl.glBegin(bgl.GL_LINE_STRIP) 
        for x,y in cur:
            bgl.glVertex2f(x,y)
        bgl.glEnd()  
    else:
        font_id = 0
        blf.size(font_id, 20, 72)
        blf.position(font_id, uploc[0], uploc[1], 0)
        blf.draw(font_id, lenx)       
        blf.position(font_id, downloc[0], downloc[1], 0)
        blf.draw(font_id, leny)
    
                     
    # restore opengl defaults
    bgl.glLineWidth(1)
    bgl.glDisable(bgl.GL_BLEND)
    bgl.glColor4f(0.0, 0.0, 0.0, 1.0)

# Defining the second graphical set that will show height during extrusion:
# Unfortunately, this doesn't work yet:
def draw_callback2_px(self, context):
    
    font_id = 0
    # First color is for lighter themes, second for default and darker themes:    
    #bgl.glColor4f(1.0, 1.0, 1.0, 1.0)
    #bgl.glColor4f(1.0, 1.0, 1.0, 0.8)   
    #blf.size(font_id, 20, 72)   
    #blf.position(font_id, 15, 30, 0)
    #blf.draw(font_id, str(Storage.z))      
    # restore opengl defaults
    bgl.glLineWidth(1)
    bgl.glDisable(bgl.GL_BLEND)
    bgl.glColor4f(0.0, 0.0, 0.0, 1.0)

# Defining the operator that will use the graphical sets and make the drawing of the base:
class NPFBDrawBase(bpy.types.Operator):
    bl_idname = "np.fb_draw_base"
    bl_label = "NP FB Draw Base"
    bl_options={'INTERNAL'}     

    def modal(self, context, event):
        context.area.tag_redraw()

        if event.type == 'MOUSEMOVE':
            self.co2d = ((event.mouse_region_x, event.mouse_region_y))
            self.mode = 1
                        
        elif event.type == 'LEFTMOUSE' and event.value == 'PRESS':
            self.points2d.append((event.mouse_region_x, event.mouse_region_y))                        
            self.mode = 2          
            if len(self.points2d) == 1:
                self.co2d = ((event.mouse_region_x, event.mouse_region_y))                                
            if len(self.points2d) == 2:
                Storage.base = self.base                             
                bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
                return {'FINISHED'}
        
        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            return {'CANCELLED'}
        
        elif event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        if context.area.type == 'VIEW_3D':
            args = (self, context)
            self._handle = bpy.types.SpaceView3D.draw_handler_add(draw_callback1_px, args, 'WINDOW', 'POST_PIXEL')

            self.trail2d = []
            self.trail = []            
            self.points2d = []
            self.points = []
            self.up2d = []
            self.up = [(0,0,0), (0,0,0), (0,0,0)]
            self.down2d = []
            self.down = [(0,0,0), (0,0,0), (0,0,0)]
            self.base = [(0,0,0), (0,0,0), (0,0,0), (0,0,0)]         
            
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "View3D not found, cannot run operator")
            return {'CANCELLED'}
        
# Defining the operator that will generate the base base mesh from the given coordinates and prepare it for extrusion:
class NPFBMakeBase(bpy.types.Operator):
    bl_idname = 'np.fb_make_base'
    bl_label = 'NP FB Make Base'
    bl_options = {'INTERNAL'}    
    
    def execute(self, context):
        # First, storing all of the system preferences set by the user, that will be changed during the process, in order to restore them when the operation is completed:
        Storage.use_snap = bpy.context.tool_settings.use_snap 
        Storage.snap_element = bpy.context.tool_settings.snap_element
        Storage.snap_target = bpy.context.tool_settings.snap_target
        Storage.pivot_point = bpy.context.space_data.pivot_point
        Storage.trans_orient = bpy.context.space_data.transform_orientation
        Storage.mesh_mode = bpy.context.tool_settings.mesh_select_mode      
        # Reading and storing the selection:
        selob = bpy.context.selected_objects
        Storage.selob = selob        
        baseverts = Storage.base
        baseedges = []
        i = 0
        for ve in baseverts:
            if i == len(baseverts) - 1:
                baseedges.append((i, 0))
            else:
                baseedges.append((i, i+1))
            i = i+1 
        basefaces = []
        baseme = bpy.data.meshes.new('float_box')
        baseme.from_pydata(baseverts,baseedges,basefaces)
        baseob = bpy.data.objects.new('float_box',baseme)
        baseob.location = mathutils.Vector((0,0,0))
        scn = bpy.context.scene
        scn.objects.link(baseob)
        scn.objects.active = baseob
        scn.update()
        bpy.ops.object.select_all(action = 'DESELECT')
        baseob.select = True                
        bpy.ops.object.origin_set(type = 'ORIGIN_GEOMETRY')
        activelayer = bpy.context.scene.active_layer
        print ('activelayer:', activelayer)
        layers = [False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
        layers[activelayer] = True
        layers = tuple(layers)
        print(layers)
        bpy.ops.object.move_to_layer(layers = layers)
        Storage.baseob = baseob
        bpy.context.scene.objects.active = baseob
        bpy.ops.object.mode_set(mode = 'EDIT')
        bpy.context.tool_settings.mesh_select_mode = True, False, False
        bpy.context.tool_settings.snap_element = 'VERTEX'
        bpy.context.tool_settings.snap_target = 'CLOSEST'
        bpy.ops.mesh.select_all(action = 'SELECT')
        bpy.ops.mesh.edge_face_add()                                             
        #print('030') 
        return{'FINISHED'}

# Defining the operator that will let the user extrude the base mesh:        
class NPFBExtrude(bpy.types.Operator):
    bl_idname = "np.fb_extrude"
    bl_label = "NP FB Extrude"
    bl_options = {'INTERNAL'} 
    
    count=0   
    
    def modal(self, context, event):
        self.count+=1
        
        # Some code that didn't succede to give the info on the extrusion height 
        # and mouse position durring this modal process:
        #selob = Storage.selob
        #self.culoc = ((mouse_region_x, mouse_region_y))
                
        if self.count == 1:
            bpy.ops.view3d.edit_mesh_extrude_move_normal('INVOKE_DEFAULT')
            
            # Also usuccesfull code:
            baseob = Storage.baseob
            me = baseob.data
            for me in me.vertices:
                z = me.co[2]
            Storage.z = z 
        
        # Not to mention, unsuccesfull:              
        #elif event.type == 'MOUSEMOVE':
            #print('mousemove')
             
        elif event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'} and event.value=='RELEASE':
            return {'PASS_THROUGH'}
        
        elif event.type in ('LEFTMOUSE','RET','NUMPAD_ENTER','SPACE') and event.value=='RELEASE':
            # Primary closing of the object and cleanup:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            bpy.ops.mesh.select_all(action='SELECT')
            bpy.ops.mesh.normals_make_consistent(inside=False)
            bpy.ops.mesh.select_all(action='DESELECT')
            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')         
            bpy.context.tool_settings.use_snap=Storage.use_snap
            bpy.context.tool_settings.snap_element=Storage.snap_element
            bpy.context.tool_settings.snap_target=Storage.snap_target
            bpy.context.space_data.pivot_point=Storage.pivot_point
            bpy.context.space_data.transform_orientation=Storage.trans_orient
            bpy.context.tool_settings.mesh_select_mode = Storage.mesh_mode 
            #print('080')                        
            return{'FINISHED'}
        
        elif event.type == 'RIGHTMOUSE':
            # Primary closing of the object and cleanup if the user decides he only wants a plane:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            bpy.ops.transform.translate(value=(0.0,0.0,0.001))            
            bpy.ops.mesh.select_all(action='INVERT')
            bpy.ops.mesh.normals_make_consistent(inside=False)
            bpy.ops.mesh.flip_normals()
            bpy.ops.mesh.select_all(action='INVERT')
            bpy.ops.mesh.delete(type='VERT')                       
            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')         
            bpy.context.tool_settings.use_snap=Storage.use_snap
            bpy.context.tool_settings.snap_element=Storage.snap_element
            bpy.context.tool_settings.snap_target=Storage.snap_target
            bpy.context.space_data.pivot_point=Storage.pivot_point
            bpy.context.space_data.transform_orientation=Storage.trans_orient
            bpy.context.tool_settings.mesh_select_mode = Storage.mesh_mode
            return{'FINISHED'}
        
        elif event.type == 'ESC':
            # Primary cleanup and deletion if the user decides to anull the whole process:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')                      
            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.ops.object.delete('EXEC_DEFAULT') 
            bpy.ops.object.select_all(action='DESELECT')
            for ob in selob:
                ob.select=True
            bpy.context.scene.objects.active=selob[len(selob)-1]
            bpy.context.tool_settings.use_snap=Storage.use_snap
            bpy.context.tool_settings.snap_element=Storage.snap_element
            bpy.context.tool_settings.snap_target=Storage.snap_target
            bpy.context.space_data.pivot_point=Storage.pivot_point
            bpy.context.space_data.transform_orientation=Storage.trans_orient
            bpy.context.tool_settings.mesh_select_mode = Storage.mesh_mode           
            return{'CANCELLED'}
                
        return{'PASS_THROUGH'}

    def invoke(self,context,event):
        #print("START_____")
        args=(self,context)
        self._handle=bpy.types.SpaceView3D.draw_handler_add(draw_callback2_px,args,'WINDOW','POST_PIXEL')        
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}
        
# Defining the operator that will do the last cleanup and bring things back to normal:
class NPFBCleanup(bpy.types.Operator):
    bl_idname = 'np.fb_cleanup'
    bl_label = 'NP FB Cleanup'
    bl_options = {'INTERNAL'}    
    
    def execute(self, context):
        if bpy.context.mode == 'EDIT':
            bpy.ops.mesh.select_all(action = 'SELECT')
            bpy.ops.mesh.normals_make_consistent(inside = False)
            bpy.ops.mesh.select_all(action = 'DESELECT')
        bpy.ops.object.mode_set(mode = 'OBJECT')
        bpy.ops.object.origin_set(type = 'ORIGIN_GEOMETRY')         
        bpy.context.tool_settings.use_snap = Storage.use_snap
        bpy.context.tool_settings.snap_element = Storage.snap_element
        bpy.context.tool_settings.snap_target = Storage.snap_target
        bpy.context.space_data.pivot_point = Storage.pivot_point
        bpy.context.space_data.transform_orientation = Storage.trans_orient
        bpy.context.tool_settings.mesh_select_mode = Storage.mesh_mode
        return{'FINISHED'}        

# This is the actual addon process, the algorithm that defines the order of operator activation inside the main macro:           
def register():
    bpy.utils.register_module(__name__)
    
    NPFloatBox007.define("NP_OT_fb_draw_base")
    NPFloatBox007.define("NP_OT_fb_make_base")
    NPFloatBox007.define("NP_OT_fb_extrude")
    NPFloatBox007.define("NP_OT_fb_cleanup")

def unregister():
  bpy.utils.unregister_module(__name__)
       
if __name__ == "__main__":
  register()