
# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# Many thanks to Blender Artists, Mano Wii and Elegant Artz.

bl_info = {
  'name':'NP Float Cone 007',
  'author':'Okavango and the ba community',
  'version':(0,0,7),
  'blender':(2,75,0),
  'location':'View3D',
  'description':'Create a cone or a pyramid with three clicks',
  'category':'3D View'}

import bpy
import mathutils, bmesh, math
from math import sin, cos, tan, atan, degrees, radians, asin, acos
import bgl
import blf
from mathutils import Vector, Matrix
from bpy_extras import view3d_utils

# Defining the main class - the macro:
class NPFloatCone007(bpy.types.Macro):
    bl_idname = 'np.float_cone_007'
    bl_label = 'NP Float Cone 007'
    bl_options = {'REGISTER','UNDO'}
    
# Defining the storage class that will serve as a varable-bank for exchange among the classes. 
#Later, this bank will recieve more variables with their values for safe keeping, as the program goes on:
class Storage:
    sides = 32  
    z = 1.0
    use_snap = None    
    snap_element = None
    snap_target = None

# Defining the first graphical set that will be shown during first phase and updated along:    
def draw_callback1_px(self, context):    
    print('trail'+ str(len(self.trail)))
        
    # Calculating the 3d points where cursor ray hits the ground:
    region = context.region
    rv3d = context.region_data
    co2d = self.co2d   
    view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, co2d)
    ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, co2d)
    ray_target = ray_origin + view_vector

    center = Vector((0,0,0))
    normal = Vector((0.0, 0.0, 1.0))
    co3dv = mathutils.geometry.intersect_line_plane(ray_origin, ray_target, center, normal, False)
    co3d = co3dv.to_tuple(4)    
    print('co3d' ,co3d)

    
    # Storring the 3d points into lists for drawing the base: 
    if self.mode == 1:
        self.trail.append((co3d))
        print('1')
    if self.mode == 2:
        self.points.append((co3d))
        self.rad.append((co3d))
        self.rad.append((co3d))
        self.cen = self.rad[0]
        print('2')
    # Correcting some persistent mistake that appears, for which i don't know the reason:
    if len(self.points) > 1:
        a = [(0,0,0)]
        a[0] = self.points[0]
        self.points = a    
    if len(self.rad) > 2:
        self.rad = [co3d, co3d]
        print('oops!')                           
    #These prints are my control in the process of writing:
    print('self.points', self.points)
    print('self.rad', self.rad)    

    # Adjusting the lists for the base while the mouse is still moving:
    j = len(self.trail)    
    if len(self.points) == 1:        
        self.rad[1] = co3d
        print('self.rad', self.rad)
    
        c = self.rad[0]
        a = self.rad[1]
        cv = Vector(c)
        av = Vector(a)
        lenrad = (av-cv).length
    
        self.base = [(self.rad[1])]
        sides = Storage.sides
        angstep = 360 / sides
        cosalpha = (a[0] - c[0]) / lenrad
        sinalpha = (a[1] - c[1]) / lenrad
        if sinalpha < 0:
            alpha = 360 - round(degrees(acos(cosalpha)), 2)
        else:
            alpha = round(degrees(acos(cosalpha)), 2)
        #print('alpha', alpha)
        beta = alpha
        i = 0       
        while i < sides:
            #print('i', i)
            beta = beta + angstep
            cosbeta = round(cos(radians(beta)), 8)
            xb = cosbeta * lenrad + c[0]
            xb = round(xb, 4)
            sinbeta = round(sin(radians(beta)), 8)
            yb = sinbeta * lenrad + c[1]
            yb = round(yb, 4)               
            self.base.append((xb, yb, 0))
            i = i+1
           
        # Before they get 2d flattened, recording the lists in 3d for the latter mesh generation:

        print('self.base', self.base)
        c2d = view3d_utils.location_3d_to_region_2d(region, rv3d, c)
        c2d = c2d.to_tuple(0)
        base2dlist = [(c2d)]  
        # Drawing the base in 2d on the screen, from 3d points:
        bgl.glEnable(bgl.GL_BLEND)
        # First color is for lighter themes, second for default and darker themes:       
        #bgl.glColor4f(1.0, 1.0, 1.0, 1.0)
        bgl.glColor4f(1.0, 1.0, 1.0, 0.6)
        bgl.glLineWidth(1)    
        bgl.glBegin(bgl.GL_LINE_STRIP)
        bgl.glVertex2f(*c2d)
        for co in self.base:
            print('co', co)
            base2d = view3d_utils.location_3d_to_region_2d(region, rv3d, co)
            bgl.glVertex2f(*base2d)
            base2dlist.append(base2d.to_tuple(2))
        self.base.remove(self.base[len(self.base) - 1])
        print('base2dlist', base2dlist)      
        bgl.glEnd()
        
        # Filling base with a transparent color:
        # First color is for lighter themes, second for default and darker themes:      
        #bgl.glColor4f(1.0, 1.0, 1.0, 0.3)
        bgl.glColor4f(1.0, 1.0, 1.0, 0.2)        
        bgl.glBegin(bgl.GL_TRIANGLE_FAN)
        for co in base2dlist:
            bgl.glVertex2f(*co)
        bgl.glEnd() 
        
        # Calculating the dimensions on the sides of the base:
        radloc = []
        radloc.append((base2dlist[0][0]+base2dlist[1][0])/2)
        radloc.append((base2dlist[0][1]+base2dlist[1][1])/2)      
        lenrad = str(abs(round(lenrad, 2)))
    
   
    

    
    # Drawing the small icon near the cursor and the dimensions:
    # First color is for lighter themes, second for default and darker themes:
    #bgl.glColor4f(1.0, 1.0, 1.0, 1.0)
    bgl.glColor4f(1.0, 1.0, 1.0, 0.8)        
    if self.base == []:
        cur = [[21, 31], [19, 30], [17, 30], [15, 31], [18, 36], [21, 31]]    
        scale = 2.0
        for co in cur:    
            co[0] = round((co[0] * scale),0) -13 + co2d[0]
            co[1] = round((co[1] * scale),0) -65 + co2d[1]        
        bgl.glBegin(bgl.GL_LINE_STRIP) 
        for x,y in cur:
            bgl.glVertex2f(x,y)
        bgl.glEnd()  
    else:
        quart = int(round((sides/4),0))
        quart = base2dlist[quart+1]
        font_id = 0
        blf.size(font_id, 20, 72)
        blf.position(font_id, radloc[0], radloc[1], 0)
        blf.draw(font_id, lenrad)
        blf.position(font_id, quart[0], quart[1] + 5, 0)
        blf.draw(font_id, str(sides))               

                            
    # restore opengl defaults
    bgl.glLineWidth(1)
    bgl.glDisable(bgl.GL_BLEND)
    bgl.glColor4f(0.0, 0.0, 0.0, 1.0)

# Defining the second graphical set that will show height during extrusion:
# Unfortunately, this doesn't work yet:
def draw_callback2_px(self, context):
    
    font_id = 0
    # First color is for lighter themes, second for default and darker themes:    
    #bgl.glColor4f(1.0, 1.0, 1.0, 1.0)
    #bgl.glColor4f(1.0, 1.0, 1.0, 0.8)   
    #blf.size(font_id, 20, 72)   
    #blf.position(font_id, 15, 30, 0)
    #blf.draw(font_id, str(Storage.z))      
    # restore opengl defaults
    bgl.glLineWidth(1)
    bgl.glDisable(bgl.GL_BLEND)
    bgl.glColor4f(0.0, 0.0, 0.0, 1.0)

# Defining the operator that will use the graphical sets and make the drawing of the base base:
class NPFCDrawBase(bpy.types.Operator):
    bl_idname = "np.fc_draw_base"
    bl_label = "NP FC Draw Base"
    bl_options={'INTERNAL'}     

    def modal(self, context, event):
        context.area.tag_redraw()

        if event.type == 'MOUSEMOVE':
            self.co2d = ((event.mouse_region_x, event.mouse_region_y))
            self.mode = 1
            
        elif event.ctrl and event.type == 'WHEELDOWNMOUSE':
            sides = Storage.sides
            sides = sides + 1
            Storage.sides = sides
            
        elif event.ctrl and event.type == 'WHEELUPMOUSE':
            sides = Storage.sides
            if sides > 3:
                sides = sides - 1
            Storage.sides = sides                            
            
        elif event.type in {'NUMPAD_PLUS', 'UP_ARROW'} and event.value == 'PRESS':
            sides = Storage.sides
            sides = sides + 1
            Storage.sides = sides       
            
        elif event.type in {'NUMPAD_MINUS', 'DOWN_ARROW'} and event.value == 'PRESS':
            sides = Storage.sides
            if sides > 3:
                sides = sides - 1
            Storage.sides = sides                                                 
                        
        elif event.type == 'LEFTMOUSE' and event.value == 'PRESS':
            self.points2d.append((event.mouse_region_x, event.mouse_region_y))            
            self.mode = 2          
            if len(self.points2d) == 1:
                self.co2d = ((event.mouse_region_x, event.mouse_region_y))                
            if len(self.points2d) == 2:
                Storage.base = self.base              
                bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
                return {'FINISHED'}
        
        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            return {'CANCELLED'}
        
        elif event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}
        
        #elif event.type in {'WHEELUPMOUSE', 'WHEELDOWNMOUSE'} and event.type not in {'LEFT_CTRL'}:
            #return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        if context.area.type == 'VIEW_3D':
            args = (self, context)
            self._handle = bpy.types.SpaceView3D.draw_handler_add(draw_callback1_px, args, 'WINDOW', 'POST_PIXEL')

            self.trail2d = []
            self.trail = []            
            self.points2d = []
            self.points = []
            self.up2d = []
            self.up = [(0,0,0), (0,0,0), (0,0,0)]
            self.down2d = []
            self.down = [(0,0,0), (0,0,0), (0,0,0)]
            self.base = []
            self.rad = []
            self.base = []         
            
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "View3D not found, cannot run operator")
            return {'CANCELLED'}
        
# Defining the operator that will generate the base mesh from the given coordinates and prepare it for extrusion:
class NPFCMakeBase(bpy.types.Operator):
    bl_idname = 'np.fc_make_base'
    bl_label = 'NP FC Make Base'
    bl_options = {'INTERNAL'}    
    
    def execute(self, context):
        # First, storing all of the system preferences set by the user, that will be changed during the process, in order to restore them when the operation is completed:
        Storage.use_snap = bpy.context.tool_settings.use_snap 
        Storage.snap_element = bpy.context.tool_settings.snap_element
        Storage.snap_target = bpy.context.tool_settings.snap_target
        Storage.pivot_point = bpy.context.space_data.pivot_point
        Storage.trans_orient = bpy.context.space_data.transform_orientation 
        Storage.mesh_mode = bpy.context.tool_settings.mesh_select_mode
        print(list(Storage.mesh_mode))    
        # Reading and storing the selection:
        selob = bpy.context.selected_objects
        Storage.selob = selob        
        baseverts = Storage.base
        baseedges = []
        i = 0
        print(baseverts)
        for ve in baseverts:
            if i == len(baseverts) - 1:
                baseedges.append((i, 0))
            else:
                baseedges.append((i, i+1))
            i = i+1    
        basefaces = []
        baseme = bpy.data.meshes.new('float_cone')
        baseme.from_pydata(baseverts,baseedges,basefaces)
        baseob = bpy.data.objects.new('float_cone',baseme)
        baseob.location = mathutils.Vector((0,0,0))
        scn = bpy.context.scene
        scn.objects.link(baseob)
        scn.objects.active = baseob
        scn.update()
        bpy.ops.object.select_all(action = 'DESELECT')
        baseob.select = True                
        bpy.ops.object.origin_set(type = 'ORIGIN_GEOMETRY')
        activelayer = bpy.context.scene.active_layer
        print ('activelayer:', activelayer)
        layers = [False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
        layers[activelayer] = True
        layers = tuple(layers)
        print(layers)
        bpy.ops.object.move_to_layer(layers = layers)        
        Storage.baseob = baseob
        bpy.context.scene.objects.active = baseob
        bpy.ops.object.mode_set(mode = 'EDIT')
        print(list(Storage.mesh_mode))
        bpy.context.tool_settings.mesh_select_mode = True, False, False
        print(list(Storage.mesh_mode))
        bpy.context.tool_settings.snap_element = 'VERTEX'
        bpy.context.tool_settings.snap_target = 'CLOSEST'
        bpy.ops.mesh.select_all(action = 'SELECT')
        bpy.ops.mesh.edge_face_add()
        bpy.ops.mesh.extrude_region_move()
        bpy.ops.mesh.merge(type = 'CENTER')                                             
        #print('030') 
        return{'FINISHED'}

# Defining the operator that will let the user extrude the base mesh:        
class NPFCExtrude(bpy.types.Operator):
    bl_idname = "np.fc_extrude"
    bl_label = "NP FC Extrude"
    bl_options = {'INTERNAL'} 
    
    count=0   
    
    def modal(self, context, event):
        self.count+=1
        
        # Some code that didn't succede to give the info on the extrusion height 
        # and mouse position durring this modal process:
        #selob = Storage.selob
        #self.culoc = ((mouse_region_x, mouse_region_y))
                
        if self.count == 1:
            
            
            bpy.ops.transform.translate('INVOKE_DEFAULT', constraint_axis = (False, False, True))
            
            # Also usuccesfull code:
            baseob = Storage.baseob
            me = baseob.data
            for me in me.vertices:
                z = me.co[2]
            Storage.z = z 
        
        # Not to mention, unsuccesfull:              
        #elif event.type == 'MOUSEMOVE':
            #print('mousemove')
             
        elif event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'} and event.value=='RELEASE':
            return {'PASS_THROUGH'}
        
        elif event.type in ('LEFTMOUSE','RET','NUMPAD_ENTER','SPACE') and event.value=='RELEASE':
            # Primary closing of the object and cleanup:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            bpy.ops.mesh.select_all(action = 'SELECT')
            bpy.ops.mesh.normals_make_consistent(inside = False)
            bpy.ops.mesh.select_all(action = 'DESELECT')
            bpy.context.tool_settings.mesh_select_mode = Storage.mesh_mode
            bpy.ops.object.mode_set(mode = 'OBJECT')
            bpy.ops.object.origin_set(type = 'ORIGIN_GEOMETRY')         
            bpy.context.tool_settings.use_snap = Storage.use_snap
            bpy.context.tool_settings.snap_element = Storage.snap_element
            bpy.context.tool_settings.snap_target = Storage.snap_target
            bpy.context.space_data.pivot_point = Storage.pivot_point
            bpy.context.space_data.transform_orientation = Storage.trans_orient 
            print(list(Storage.mesh_mode))
            #print('080')                        
            return{'FINISHED'}
        
        elif event.type == 'RIGHTMOUSE':
            # Primary closing of the object and cleanup if the user decides he only wants a plane:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            bpy.ops.transform.translate(value=(0.0,0.0,0.001))            
            bpy.ops.mesh.select_all(action='INVERT')
            bpy.ops.mesh.normals_make_consistent(inside=False)
            bpy.ops.mesh.flip_normals()
            bpy.ops.mesh.select_all(action='INVERT')
            bpy.ops.mesh.delete(type='VERT')                       
            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')         
            bpy.context.tool_settings.use_snap=Storage.use_snap
            bpy.context.tool_settings.snap_element=Storage.snap_element
            bpy.context.tool_settings.snap_target=Storage.snap_target
            bpy.context.space_data.pivot_point=Storage.pivot_point
            bpy.context.space_data.transform_orientation=Storage.trans_orient
            bpy.context.tool_settings.mesh_select_mode = Storage.mesh_mode
            return{'FINISHED'}
        
        elif event.type == 'ESC':
            # Primary cleanup and deletion if the user decides to anull the whole process:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')                      
            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.ops.object.delete('EXEC_DEFAULT') 
            bpy.ops.object.select_all(action='DESELECT')
            for ob in selob:
                ob.select=True
            bpy.context.scene.objects.active=selob[len(selob)-1]
            bpy.context.tool_settings.use_snap=Storage.use_snap
            bpy.context.tool_settings.snap_element=Storage.snap_element
            bpy.context.tool_settings.snap_target=Storage.snap_target
            bpy.context.space_data.pivot_point=Storage.pivot_point
            bpy.context.space_data.transform_orientation=Storage.trans_orient
            bpy.context.tool_settings.mesh_select_mode = Storage.mesh_mode           
            return{'CANCELLED'}
                
        return{'PASS_THROUGH'}

    def invoke(self,context,event):
        #print("START_____")
        args=(self,context)
        self._handle=bpy.types.SpaceView3D.draw_handler_add(draw_callback2_px,args,'WINDOW','POST_PIXEL')        
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}
        
# Defining the operator that will do the last cleanup and bring things back to normal:
class NPFCCleanup(bpy.types.Operator):
    bl_idname = 'np.fc_cleanup'
    bl_label = 'NP FC Cleanup'
    bl_options = {'INTERNAL'}    
    
    def execute(self, context):
        if bpy.context.mode == 'EDIT':
            bpy.ops.mesh.select_all(action = 'SELECT')
            bpy.ops.mesh.normals_make_consistent(inside = False)
            bpy.ops.mesh.select_all(action = 'DESELECT')
        bpy.ops.object.mode_set(mode = 'OBJECT')
        bpy.ops.object.origin_set(type = 'ORIGIN_GEOMETRY')         
        bpy.context.tool_settings.use_snap = Storage.use_snap
        bpy.context.tool_settings.snap_element = Storage.snap_element
        bpy.context.tool_settings.snap_target = Storage.snap_target
        bpy.context.space_data.pivot_point = Storage.pivot_point
        bpy.context.space_data.transform_orientation = Storage.trans_orient
        bpy.context.tool_settings.mesh_select_mode = Storage.mesh_mode
        return{'FINISHED'}        

# This is the actual addon process, the algorithm that defines the order of operator activation inside the main macro:           
def register():
    bpy.utils.register_module(__name__)
    
    NPFloatCone007.define("NP_OT_fc_draw_base")
    NPFloatCone007.define("NP_OT_fc_make_base")
    NPFloatCone007.define("NP_OT_fc_extrude")
    NPFloatCone007.define("NP_OT_fc_cleanup")

def unregister():
  bpy.utils.unregister_module(__name__)
       
if __name__ == "__main__":
  register()